package dev.fitko.fitconnect.core;

import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jwt.SignedJWT;
import dev.fitko.fitconnect.api.domain.model.event.EventLog;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.EventPayload;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.validation.ValidationContext;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.EventLogException;
import dev.fitko.fitconnect.api.services.Subscriber;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.api.services.events.EventLogService;
import dev.fitko.fitconnect.api.services.events.EventLogVerificationService;
import dev.fitko.fitconnect.api.services.events.SecurityEventService;
import dev.fitko.fitconnect.api.services.submission.SubmissionService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import dev.fitko.fitconnect.core.util.EventLogUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Collectors;

public class SubmissionSubscriber implements Subscriber {

    private static final Logger LOGGER = LoggerFactory.getLogger(SubmissionSubscriber.class);

    private final SubmissionService submissionService;
    private final CryptoService cryptoService;
    private final ValidationService validationService;
    private final EventLogService eventLogService;

    private final EventLogVerificationService eventLogVerifier;
    private final SecurityEventService securityEventService;

    public SubmissionSubscriber(final SubmissionService submissionService,
                                final EventLogService eventLogService,
                                final EventLogVerificationService eventLogVerifier,
                                final CryptoService cryptoService,
                                final ValidationService validationService,
                                final SecurityEventService securityEventService) {
        this.submissionService = submissionService;
        this.eventLogService = eventLogService;
        this.eventLogVerifier = eventLogVerifier;
        this.cryptoService = cryptoService;
        this.validationService = validationService;
        this.securityEventService = securityEventService;
    }

    @Override
    public byte[] decryptStringContent(final RSAKey privateKey, final String encryptedContent) {
        return this.cryptoService.decryptBytes(privateKey, encryptedContent);
    }

    @Override
    public Set<SubmissionForPickup> pollAvailableSubmissions(final UUID destinationId, final int offset, final int limit) {
        return this.submissionService.pollAvailableSubmissions(destinationId, offset, limit).getSubmissions();
    }

    @Override
    public Submission getSubmission(final UUID submissionId) {
        return this.submissionService.getSubmission(submissionId);
    }

    @Override
    public String fetchAttachment(final UUID submissionId, final UUID attachmentId) {
        return this.submissionService.getAttachment(submissionId, attachmentId);
    }

    @Override
    public List<EventLogEntry> getEventLog( final UUID caseId, final UUID destinationId) {
        final EventLog eventLog = this.eventLogService.getEventLog(caseId);
        final List<SignedJWT> events = EventLogUtil.getJWTSFromEvents(eventLog.getEventLogs());
        final ValidationContext ctx = new ValidationContext(destinationId, caseId);
        final List<ValidationResult> results = this.eventLogVerifier.validateEventLogs(ctx, events);
        if (!results.isEmpty()) {
            final String errorMessages = results.stream()
                    .map(ValidationResult::getError)
                    .map(Throwable::getMessage)
                    .distinct()
                    .collect(Collectors.joining("\n"));
            throw new EventLogException("Event log contains invalid entries: \n" + errorMessages);
        }
        return EventLogUtil.mapEventLogToEntries(eventLog);
    }

    @Override
    public ValidationResult validateMetadata(final Metadata metadata) {
        return this.validationService.validateMetadataSchema(metadata);
    }

    @Override
    public ValidationResult validateHashIntegrity(final String originalHash, final byte[] data) {
        return this.validationService.validateHashIntegrity(originalHash, data);
    }

    @Override
    public void acceptSubmission(final EventPayload eventPayload) {
        final SignedJWT confirmedSubmissionEvent = this.securityEventService.createAcceptSubmissionEvent(eventPayload);
        this.eventLogService.sendEvent(eventPayload.getCaseId(), confirmedSubmissionEvent.serialize());
        LOGGER.info("CONFIRMED submission {} successfully", eventPayload.getSubmissionId());
    }

    @Override
    public void rejectSubmission(final EventPayload eventPayload) {
        final SignedJWT rejectSubmissionEvent = this.securityEventService.createRejectSubmissionEvent(eventPayload);
        this.eventLogService.sendEvent(eventPayload.getCaseId(), rejectSubmissionEvent.serialize());
        LOGGER.info("REJECTED submission {}", eventPayload.getSubmissionId());
    }
}
